%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 1999
%%%
%%% Last change:
%%%   $Date: 2000/12/21 11:59:11 $ by $Author: afranke $
%%%   $Revision: 1.2 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% generate the gump parser source code
%%

functor
   
import
   DtdRules at 'DtdRules.ozf'
   
export
   writeSyn:    WriteRules   % fun{$ Elements SemIsStatic} --> VS 
   writeTok:    WriteTokens  % fun{$ Elements} --> VS
   writeSpec:   WriteSpec    % fun{$ Elements} --> VS
   writeParser: WriteParser 
   % fun{$ SemUrl XmlTok XmlSyn ElTok ElSyn ElSpec StartElems} --> VS
   
   filePrefix:  FilePrefix   % VS
   
define
   
   FilePrefix =
   "%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %%%"#"\n"#
   "%%%  This is a generated file!  Do not edit.  %%%"#"\n"#
   "%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% %%%"#"\n"#
   "\n"

   MethPrefix   = "make_"
   SemClassName = "Sem"
   
   Tok  = DtdRules.makeTokenName
   /*
   Aux  = DtdRules.makeAuxName
   Elem = DtdRules.makeElemName
   Term = DtdRules.makeTerminalName
   */
   
   fun {WriteRules Elements SemIsStatic}
      SemClass    = if SemIsStatic\=false then SemClassName else false end
      WriteAction = {DtdRules.makeWriteAction SemClass MethPrefix}
   in
      {DtdRules.writeElementRules Elements WriteAction}#"\n"#
      {DtdRules.writeSpecialRule "START_" Elements}
   end
   
   local
      fun {DoEnumerate Xs Sep F}
         case Xs of nil then ''
         [] X|Xr then Sep#{F X}#{DoEnumerate Xr Sep F} end
      end
   in
      fun {Enumerate Xs Sep F}
         case Xs of nil then ''
         [] X|Xr then {F X}#{DoEnumerate Xr Sep F} end
      end
   end
   local
      fun {Id X} X end
   in
      fun {Enum Xs Sep}
         {Enumerate Xs Sep Id}
      end
   end

   %%
   %% tokens, specification
   %%
   local
      fun {WriteAtom A}
         "'"#{AtomToString A}#"'"
      end
      fun {WriteName E}
         {WriteAtom E.name}
      end
      %% does the element content contain text?
      fun {WriteBool Bool}
         if Bool
         then "true"
         else "false"
         end
      end   
      fun {IsText E}
         case E.content
         of 'PCDATA' then true
         [] 'ANY'    then true
         [] mixed(_) then true
         else false
         end
      end
      %% attribute default values
      fun {WriteAttDefaults E}
         case {CondSelect E attlist unit} of unit then {WriteName E}
         [] AttList then
            Defaults = {Filter
                        {Map AttList
                         fun {$ N#(_#DefaultDecl)}
                            case DefaultDecl
                            of '#REQUIRED'   then false
                            [] '#IMPLIED'    then false
                            [] '#FIXED'(Val) then N#Val
                            [] unit(Val)     then N#Val
                            end
                         end}
                        fun {$ X} X\=false end}
         in
            if Defaults == nil then {WriteName E}
            else
               fun {WriteAttVal N#Val}
                  {WriteAtom N}#":"#'"'#Val#'"'
               end
            in
               {WriteName E}#"("#{Enumerate Defaults " " WriteAttVal}#")"
            end
         end
      end
   in
      fun {WriteTokens Elements}
         {Enumerate Elements "\n" fun {$ E} {Tok E.name} end}#"\n" 
      end
      /*
      fun {WriteTokenSpec Elements}
         "tokens("#"\n"#{Enumerate Elements "\n" TokenLine}#"\n"#")"#"\n"
      end
      fun {WriteTextSpec Elements}
         "text("#"\n"#{Enumerate Elements "\n" TextLine}#"\n"#")"#"\n"
      end
      fun {WriteAttDefaults Elements}
         "defaults("#"\n"#{Enumerate Elements "\n" DefaultsLine}#"\n"#")"#"\n"
      end
      */
      fun {WriteSpecLine E}
         {WriteName E}
         #"("
         #"'token':"#{Tok E.name}#" "
         #"text:"#{WriteBool {IsText E}}#" "
         #"defaults:"#{WriteAttDefaults E}
         #")"
      end
      fun {WriteSpec Elements}
         {Enumerate Elements "\n" WriteSpecLine}#"\n"
      end
   end

   %%
   %% parser
   %%
   local
      fun {Concat L}
         {FoldR L Append nil}
      end
   in
      fun {WriteParser SemUrl XmlTok XmlSyn ElTok ElSyn ElSpec StartElems}
         %% Semantics = "omobj.sem.ozf"
         %% XmlTok    = "xml.tok.ozg"
         %% XmlSyn    = "xml.syn.ozg"
         %% ElTok     = "omobj.tok.ozg"
         %% ElSyn     = "omobj.syn.ozg"
         %% ElSpec    = "omobj.spec.oz"
         Imports # Parents # IsGeneric =
         if SemUrl==false then '' # '' # "true"
         else
            ("X('class': "#SemClassName#") at '"#SemUrl#"'") #
            SemClassName #
            "false"
         end

         Lines = 
         [[
           "%%%"
           "%%% gump parser definition"
           "%%%"
           ""
           "\\switch +gump"
           "%\\switch +gumpparserverbose"
           ""
           "declare F in F = "
           ""
           "functor"
           ""
           "import"
           "   GumpParser"
           "   "#Imports
           ""
           "export"
           "   'class':   Parser"
           "   elements:  Spec"
           "   isGeneric: IsGeneric"
           ""
           "prepare"
           "   SpecList = ["
           "               \\insert '"#ElSpec#"'"
           "              ]"
           "   Token    = {Map SpecList"
           "               fun {$ E} L={Label E} in L#E.'token' end}"
           "   Text     = {Map SpecList"
           "               fun {$ E} L={Label E} in L#E.text end}"
           "   Defaults = {Map SpecList"
           "               fun {$ E}"
           "                  L={Label E} in"
           "                  L#{Record.map E.defaults ByteString.make}"
           "               end}"
           ""
           "   Spec = spec('token':  {Record.adjoinList unit Token}"
           "               text:     {Record.adjoinList unit Text}"
           "               defaults: {Record.adjoinList unit Defaults}"
           "              )"
           ""
           "define"
           ""
           "   parser Parser"
           "      from"
           "         GumpParser.'class'"
           "         "#Parents
           ""
           "      token"
           "         \\insert '"#XmlTok#"'"
           "         \\insert '"#ElTok#"'"
           ""
          ]
          {Map StartElems DtdRules.writeTerminalRule}
          [
           "      syn START($)"
           "         START_xml($)"
           "      [] START_($)"
           "      end"
           ""
           "      \\insert '"#XmlSyn#"'"
           "      \\insert '"#ElSyn#"'"  
           "   end"
           ""
           "   IsGeneric = "#IsGeneric
           ""
           "end"
           ""
          ]]
      in
         {Enum {Concat Lines} "\n"}
      end
   end      
end











