%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 1999
%%%
%%% Last change:
%%%   $Date: 2000/01/16 06:08:14 $ by $Author: afranke $
%%%   $Revision: 1.1 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% DtdGraph.oz
%%
%% extract the "has-child-node" relation from a DTD
%%

functor

export
   makeGraph: MakeGraph % fun {$ Rec} --> children(<Elem>:ChildrenList...)
   
define
   
   %%
   %% collect child elements
   %%
   local
      local
         proc {PutChild D X}
            E = {VirtualString.toAtom X} in
            {Dictionary.put D E unit}
         end
         proc {CollectChildrenInList D Xs}
            case Xs of nil then skip
            [] X|Xr then
               {CollectChildren D X}
               {CollectChildrenInList D Xr}
            end
         end
      in
         proc {CollectChildren D X}
            if {IsVirtualString X} then {PutChild D X}
            elsecase X
            of '+'(Y) then {CollectChildren D Y}
            [] '*'(Y) then {CollectChildren D Y}
            [] '?'(Y) then {CollectChildren D Y}
            [] 'seq'(L)    then {CollectChildrenInList D L}
            [] 'choice'(L) then {CollectChildrenInList D L}
            end
         end
      end
   in
      fun {GetChildren M}
         case M
         of 'EMPTY'      then nil
         [] 'ANY'        then nil
         [] 'PCDATA'     then nil
         [] mixed(Elems) then {Map Elems VirtualString.toAtom} 
         [] children(Spec) then
            D = {NewDictionary} in
            {CollectChildren D Spec}
            {Dictionary.keys D}
         end
      end
   end
   
   %%
   %% main 
   %%      
   fun {MakeGraph R}
      {Adjoin
       {Record.map R.elements
        fun {$ E} {GetChildren E.content} end}
       children}
   end
end
