%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 1999
%%%
%%% Last change:
%%%   $Date: 2001/11/22 20:56:42 $ by $Author: afranke $
%%%   $Revision: 1.3 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% DtdInput.oz
%%
%% read XML Document Type Definitions
%%

functor

import
   XmlScanner(dtd: DtdScanner)
   at 'x-ozlib://mathweb/xmlparser/XmlScanner.ozf'
   
export
   readDtd:   ReadDtd         % fun{$ Url} --> Rec
   toFunctor: RecordToFunctor % fun{$ Rec} --> F
      
define
   
   local      
      %%
      %% check element references in content models
      %%
      local
         proc {CheckElem X Es}
            if {Dictionary.member Es X}\=true then
               raise undefinedElement(X) end
            end
         end
         proc {DoCheck X Es}
            if {IsAtom X} then {CheckElem X Es}
            elsecase X
            of '+'(Y) then {DoCheck Y Es}
            [] '*'(Y) then {DoCheck Y Es}
            [] '?'(Y) then {DoCheck Y Es}
            [] 'seq'(L)    then {ForAll L proc {$ Y} {DoCheck Y Es} end}
            [] 'choice'(L) then {ForAll L proc {$ Y} {DoCheck Y Es} end}
            end
         end
      in
         proc {CheckModel Model Es}
            case Model
            of 'EMPTY'        then skip
            [] 'ANY'          then skip
            [] 'PCDATA'       then skip
            [] mixed(Elems) then
               {ForAll Elems proc {$ X} {CheckElem X Es} end}
            [] children(Spec) then
               {DoCheck Spec Es}
            end
         end
         proc {CheckModels Elems}
            {ForAll {Dictionary.items Elems}
             proc {$ E} {CheckModel E.content Elems} end}         
         end
      end
      
      %%
      %% accumulator for dtd items
      %%
      class Items
         attr
            'PEs':    unit
            'GEs':    unit
            elements: unit
            
         meth init
            'PEs'    <- {NewDictionary}
            'GEs'    <- {NewDictionary}
            elements <- {NewDictionary}
         end
         
         meth getPEs($)      @'PEs'    end
         meth getGEs($)      @'GEs'    end
         meth getElements($) @elements end
         
         meth putEntity(V ?Ref)
            case V
            of '%'(N Def)  then {Dictionary.put @'PEs' N Def} Ref='%'#N
            [] '\&'(N Def) then {Dictionary.put @'GEs' N Def} Ref='\&'#N
            end 
         end
         meth putElement(V ?Ref)
            case V of 'ELEMENT'(N Model) then
               {Dictionary.put @elements N element(name:N content:Model)}
               Ref='ELEMENT'#N
            end 
         end
         meth putAttlist(V ?Ref)
            case V of 'ATTLIST'(N Defs) then
               E = {Dictionary.get @elements N} in
               {Dictionary.put @elements N {AdjoinAt E attlist Defs}}
               Ref='ATTLIST'#N
            end
         end      
         meth put(T V ?Ref)
            case T
            of 'ENTITY'  then Items, putEntity(V ?Ref)
            [] 'ELEMENT' then Items, putElement(V ?Ref)
            [] 'ATTLIST' then Items, putAttlist(V ?Ref)
            end   
         end
      end
   in
      %%
      %% main
      %%
      fun {ReadDtd Url}
         Acc   = {New Items init}
         Refs  = {Map {DtdScanner.scanURL Url}
                  fun {$ T#V} {Acc put(T V $)} end}
         PEs   = {Acc getPEs($)}
         GEs   = {Acc getGEs($)}
         Elems = {Acc getElements($)}
         {CheckModels Elems}
      in
         dtd('%':      {Dictionary.toRecord '%'      PEs}
             '\&':     {Dictionary.toRecord '\&'     GEs}
             elements: {Dictionary.toRecord elements Elems}
             items:    Refs
            )
      end

      fun {RecordToFunctor R}
         `%`        = R.'%'
         `\&`       = R.'\&'
         `elements` = R.elements
         `items`    = R.items
      in
         functor
         export
            '%':  `%`
            '\&': `\&`
            elements: `elements` 
            items:    `items`
         end
      end
   end

end
