%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 1999
%%%
%%% Last change:
%%%   $Date: 2000/01/16 06:08:15 $ by $Author: afranke $
%%%   $Revision: 1.1 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% DtdOutput.oz
%%
%% write XML Document Type Definitions
%%

functor

import
   XVS at 'XVS.ozf'
   
export
   writeDtd: WriteDtd % fun {$ Rec} --> VS
   
define
   
   local
      %%
      %% auxiliuary functions 
      %%
      
      fun {Enumerate Xs Sep F}
         case Xs of nil then ""
         [] X|Xr then Sep#{F X}#{Enumerate Xr Sep F} end
      end
      
      fun {WriteName Name}
         Name
      end
      fun {WriteNmToken NmToken}
         NmToken
      end
      fun {WriteVal Val}
         if {Member &" {VirtualString.toString Val}}
         then "'"#Val#"'"
         else '"'#Val#'"'
         end
      end
      fun {WriteSystemLiteral SystemLiteral}
         '"'#SystemLiteral#'"'
      end
      fun {WritePubidLiteral PubidLiteral}
         '"'#PubidLiteral#'"'
      end
      
      %%
      %% ELEMENT declarations
      %%
      local
         local
            fun {WriteSeq X|Xr}
               "("#{DoWrite X}#{Enumerate Xr ", " DoWrite}#")"
            end
            fun {WriteChoice X|Xr}
               "( "#{DoWrite X}#{Enumerate Xr " | " DoWrite}#" )"
            end
            fun {DoWrite X}
               if {IsVirtualString X} then X
               elsecase X
               of '+'(Y) then {DoWrite Y}#'+'       
               [] '*'(Y) then {DoWrite Y}#'*'
               [] '?'(Y) then {DoWrite Y}#'?'
               [] 'seq'(L)    then {WriteSeq L}
               [] 'choice'(L) then {WriteChoice L}
               end
            end
         in
            fun {WriteContentModel M}
               case M
               of 'EMPTY'      then 'EMPTY'
               [] 'ANY'        then 'ANY'
               [] 'PCDATA'     then "(#PCDATA)"
               [] mixed(Elems) then  {WriteChoice "#PCDATA"|Elems}#'*' 
               [] children(Spec) then
                  case Spec
                  of 'seq'(_)    then {DoWrite Spec}
                  [] 'choice'(_) then {DoWrite Spec}
                  else
                     "("#{DoWrite Spec}#")"
                  end
               end
            end
         end
      in
         fun {WriteElement E}
            "\n"#"<!ELEMENT "#E.name#" "#{WriteContentModel E.content}#" >"
         end
      end
      
      %%
      %% ATTLIST declarations
      %%
      local
         local
            local
               fun {WriteEnum N|Nr}
                  "("#{WriteNmToken N}#{Enumerate Nr "|" WriteNmToken}#")"
               end
            in
               fun {WriteType Type}
                  case Type
                  of 'CDATA'           then 'CDATA'
                  [] 'IDREFS'          then 'IDREFS'
                  [] 'IDREF'           then 'IDREF'
                  [] 'ID'              then 'ID'
                  [] 'ENTITIES'        then 'ENTITIES'
                  [] 'ENTITY'          then 'ENTITY'
                  [] 'NMTOKENS'        then 'NMTOKENS'
                  [] 'NMTOKEN'         then 'NMTOKEN'
                  [] 'NOTATION'(Ns)    then 'NOTATION'#" "#{WriteEnum Ns}
                  [] 'ENUMERATION'(Ns) then {WriteEnum Ns}
                  end
               end
            end
            fun {WriteDefaultDecl DefaultDecl}
               case DefaultDecl
               of '#REQUIRED'   then '#REQUIRED'
               [] '#IMPLIED'    then '#IMPLIED'
               [] '#FIXED'(Val) then '#FIXED'#" "#{WriteVal Val}
               [] unit(Val)     then {WriteVal Val}
               end
            end
         in
            fun {WriteAtt Name#(Type#DefaultDecl)}
               {WriteName Name}#" "#
               {WriteType Type}#" "#
               {WriteDefaultDecl DefaultDecl}
            end
         end
      in
         fun {WriteAttlist E}
            "<!ATTLIST "#E.name#" "#{Enumerate E.attlist " " WriteAtt}#" >"
         end
      end

      %%
      %% ENTITY declarations
      %%
      local
         fun {WriteExternalID ID}
            if {HasFeature ID public} then
               'PUBLIC'#
               " "#{WritePubidLiteral ID.public}#
               " "#{WriteSystemLiteral ID.system}  
            else
               'SYSTEM'#" "#{WriteSystemLiteral ID.system}
            end
         end
         fun {WriteNdata ID}
            if {HasFeature ID ndata} then
               " "#'NDATA'#" "#{WriteName ID.ndata}
            else "" end
         end
         local
            fun {Ref2VS X}
               case X
               of '%'(N)  then '%'#N#';'
               [] '\&'(N) then '\&'#N#';'
               end
            end
         in
            fun {Val2VS Val}
               {XVS.map Val Ref2VS}
            end
         end
         fun {WritePEDef Def}
            case Def of 'VAL'(Val) then {WriteVal {Val2VS Val}}
            else {WriteExternalID Def}
            end
         end
         fun {WriteGEDef Def}
            case Def of 'VAL'(Val) then {WriteVal {Val2VS Val}}
            else
               {WriteExternalID Def}#{WriteNdata Def}
            end
         end
      in
         fun {WritePE N Def}
            "\n"#"<!ENTITY % "#N#" "#{WritePEDef Def}#" >"
         end
         fun {WriteGE N Def}
            "\n"#"<!ENTITY "#N#" "#{WriteGEDef Def}#" >"
         end
      end

   in
      %%
      %% main 
      %%
      fun {WriteDtd R}
         fun {WriteItem Kind#ID}
            case Kind
            of '%'       then {WritePE ID R.'%'.ID}
            [] '\&'      then {WriteGE ID R.'\&'.ID}
            [] 'ELEMENT' then {WriteElement R.elements.ID}
            [] 'ATTLIST' then {WriteAttlist R.elements.ID}
            end
         end
      in
         {Enumerate R.items "\n" WriteItem}#"\n"
      end
   end

end
