%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 1999
%%%
%%% Last change:
%%%   $Date: 2000/12/21 14:53:36 $ by $Author: afranke $
%%%   $Revision: 1.5 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% DtdRules.oz
%%
%% generate Gump syntax rules according to the content models given in a DTD
%%

functor

import  
   X('class': FIFO) at 'FIFO.ozf'

export
   makeTokenName:     Tok   % fun{$ ElemName} --> VS
   makeAuxName:       Aux   % fun{$ AuxName}  --> VS
   makeElemName:      Elem  % fun{$ ElemName} --> VS
   makeTerminalName:  Term  % fun{$ ElemName} --> VS
   
   makeWriteAction:   MakeWriteAction   % fun{$ PrefixVS} -> (fun{$ E C M}->VS)
   writeTerminalRule: WriteTerminalRule % fun{$ ElemName}             --> VS
   writeElementRules: WriteElementRules % fun{$ Elements WriteAction} --> VS
   writeSpecialRule:  WriteSpecialRule  % fun{$ RuleName Elements}    --> VS
   
define

   fun {ToString VS}
      case VS of nil then "nil"
      else {VirtualString.toString VS} end
   end

   %%
   %% create a VirtualString enumerating a list of items
   %%
   local
      fun {DoEnumerate Xs Sep F}
         case Xs of nil then ''
         [] X|Xr then Sep#{F X}#{DoEnumerate Xr Sep F} end
      end
   in
      fun {Enumerate Xs Sep F}
         case Xs of nil then ''
         [] X|Xr then {F X}#{DoEnumerate Xr Sep F} end
      end
   end
   local
      fun {Id X} X end
   in
      fun {Enum Xs Sep}
         {Enumerate Xs Sep Id}
      end
   end
   
   %%
   %% todo: Default Attribute Values
   %%
   
   %%
   %% general format for syntax rules
   %%
   %% {WriteAlt Vars Factors Actions}
   %% {WriteAlts [Vars#Factors#Actions...]}
   %% {WriteSyn Head Alternatives}
   %%
   local
      fun {WriteVars Vars}
         if Vars==nil then "" else
            {Enum Vars " "}#" in "#"\n"#
            "   "
         end
      end
      fun {EnumOrSkip L}
         if L==nil
         then "skip"
         else {Enum L " "}
         end
      end
   in
      fun {WriteAlt Vars Factors Actions}
         {WriteVars Vars}
         #{EnumOrSkip Factors}#"\n"#
         "   => "#"\n"#
         "      "#{EnumOrSkip Actions}#"\n"
      end
      fun {WriteAlts As}
         {Map As fun {$ Vars#Factors#Actions}
                    {WriteAlt Vars Factors Actions}
                 end}
      end
      fun {WriteSyn Head As}
         "syn"#" "#Head#"\n"#
         "   "#{Enum As "[] "}#
         "end"#"\n"
      end
   end
  
   %%
   %% rule names, rule heads
   %%
   local
      AuxPrefix      = "Aux"
      ElemPrefix     = "ELEMENT_"
      TokenPrefix    = "START_" 
      TerminalPrefix = "start_"
   in
      local
         fun {MapChar C}
            case C
            of &. then &_
            [] &- then &_
            [] &: then &_
            else C
            end
         end
      in
         fun {Elem ElemName}
            ElemPrefix#{Map {ToString ElemName} MapChar} 
         end
      end
      %%
      fun {Aux R} {VirtualString.toAtom AuxPrefix#R} end
      %%
      fun {Tok  ElemName} "'"#TokenPrefix   #{ToString ElemName}#"'" end
      fun {Term ElemName} "'"#TerminalPrefix#{ToString ElemName}#"'" end
   end
   %%
   fun {AuxHead R}
      {Aux R}#"(?C ?M)"
   end
   
   %%
   %% support for content references, optionally followed by ?/*/+
   %%
   %% C: content
   %% M: content model
   
% syn Aux1(?C ?M)
%    EL(!C) => M='EL'
% end
% syn Aux2(?C ?M)
%    Aux1(!C !M)
% end
   proc {Simple C M R ?Vars ?Factor ?Action}
      case R
      of element(E) then
         Vars   = nil
         Factor = {Elem E}#"(!"#C#")"
         Action = M#"='"#E#"'"
      [] aux(R) then
         Vars   = nil
         Factor = {Aux R}#"(!"#C#" !"#M#")"
         Action = ""     
      end
   end
   
% syn Aux3(?C ?M)
%    !C=[ EL($) ] => M=if C==nil then unit else 'EL' end
% end
% syn Aux4(?C ?M)
%    C_ M_ in
%    [ Aux1(!C_ !M_) ] => if {IsDet C_}
%                         then C=C_ M=M_
%                         else C=nil M=unit
%                         end
% end
   proc {Optional C M R ?Vars ?Factor ?Action}
      case R
      of element(E) then
         Vars   = nil
         Factor = "!"#C#"=[ "#{Elem E}#"($) ]"
         Action = M#" = if "#C#"==nil then unit else '"#E#"' end"
      [] aux(R) then
         Vars   = [C#"_" M#"_"]
         Factor = "[ "#{Aux R}#"(!"#C#"_ !"#M#"_)"#" ]"
         Action =
         "if {IsDet "#C#"_} "#
         "then "#C#"="#C#"_ "#M#"="#M#"_ "#
         "else "#C#"=nil "#M#"=unit "#
         "end"
      end
   end
   
% syn Aux5(?C ?M)
%    !C=( EL($) )* => M=list('*' 'EL')
% end
% syn Aux6(?C ?M)
%    !C=( Aux1($ _) )* => M=list('*' unit)
% end
   proc {Star C M R ?Vars ?Factor ?Action}
      case R
      of element(E) then
         Vars   = nil
         Factor = "!"#C#"=( "#{Elem E}#"($) )*"
         Action = M#"=list('*' '"#E#"')"
      [] aux(R) then
         Vars   = nil
         Factor = "!"#C#"=( "#{Aux R}#"($ _)"#" )*"
         Action = M#"=list('*' unit)"
      end
   end
   
% syn Aux7(?C ?M)
%    !C=( EL($) )+ => M=list('+' 'EL')
% end
% syn Aux8(?C ?M)
%    !C=( Aux1($ _) )+ => M=list('+' unit)
% end
   proc {Plus C M R ?Vars ?Factor ?Action}
      case R
      of element(E) then
         Vars   = nil
         Factor = "!"#C#"=( "#{Elem E}#"($) )+"
         Action = M#"=list('+' '"#E#"')"
      [] aux(R) then
         Vars   = nil
         Factor = "!"#C#"=( "#{Aux R}#"($ _)"#" )+"
         Action = M#"=list('+' unit)"
      end
   end

   %%
   %% rule generation
   %%
   local
      %%
      class Counter
         attr n
         meth init
            n<-0
         end
         meth next(?N)
            N=@n+1
            n<-N
         end
      end
      %%
      class SeqManager
         attr
            vars:    nil
            factors: nil
            actions: nil   
         meth init
            vars    <- nil
            factors <- nil
            actions <- nil
         end
         meth putVars(Vars) vars    <- {Append {Reverse Vars} @vars} end
         meth putFactor(F)  factors <- F|@factors end
         meth putAction(A)  actions <- A|@actions end
         meth getVars($)    {Reverse @vars}    end
         meth getFactors($) {Reverse @factors} end
         meth getActions($) {Reverse @actions} end
      end
      %%
      fun {FindRule Entries CModel}
         case Entries of nil then nil
         [] (R#M)|Rest then
            if CModel==M then R
            else {FindRule Rest CModel}
            end
         end
      end
   in
      %%
      %% main generator class
      %%
      class Rules
         attr
            writeAction: unit
            counter:     unit
            rules:       unit
            cmodels:     unit
         meth init(writeAction:WriteAction)
            writeAction <- WriteAction
            counter     <- {New Counter init}
            rules       <- {New FIFO init}
            cmodels     <- {NewDictionary}
         end
         meth getRules($)
            {@rules getAll($)} 
         end
         meth putRule(VS)
            {@rules put(VS)}
         end
         meth putModel(R CModel)
            {Dictionary.put @cmodels R CModel}
         end
         meth AuxName(CModel ?R)
            !R = {@counter next($)}
            Rules, putModel(R CModel)
         end
         meth CondGetAux(CModel Default $)
            case {FindRule {Dictionary.entries @cmodels} CModel}
            of nil then Default
            [] R then R
            end
         end
         %%
         meth get(CModel $) % --> element(R) or aux(R)
            if {IsVirtualString CModel} then element(CModel)
            elsecase Rules, CondGetAux(CModel unit $) of unit then
               R=Rules, CreateAux(CModel $) in aux(R)
            [] R then                          aux(R)
            end
         end
         %%
         %% create new auxiliuary rules for a given content model
         %%
         meth CreateAux(CModel ?R)
            !R = Rules, AuxName(CModel $)
            As = case CModel of 'choice'(L)
                 then Rules, CreateAlternatives(L $)
                 else Rules, CreateAlternatives([CModel] $)
                 end
         in %% note: head is fix: Rulename(?C ?M)
            Rules, putRule({WriteSyn {AuxHead R} {WriteAlts As}})
         end
         meth CreateAlternatives(L $)
            {Map L fun {$ M} Rules, CreateSeq(M $) end}
         end
         meth CreateSeq(CModel $) % --> Vars#Factors#Actions     
            case CModel of 'seq'(L) then
               case L of [SubModel]
               then Rules, SimpleSeq(SubModel $)
               else Rules, ComplexSeq(L $)
               end
            else Rules, SimpleSeq(CModel $)
            end
         end
         meth SimpleSeq(CModel $) % --> Vars#Factors#Actions
            Vars Factor Action in
            Rules, CreateSimple("C" "M" CModel ?Vars ?Factor ?Action)
            Vars#[Factor]#[Action]
         end
% syn Aux9(?C ?M)
%    C1 M1 C2 M2 in
%    EL1(C1) EL2(C2) => M1='EL1' M2='EL2' C=C1#C2 M=M1#M2
% end
% syn Aux12(?C ?M)
%    C1 M1 C2 M2 in
%    Aux1(?C1 ?M1) Aux2(?C2 ?M2) => C=C1#C2 M=M1#M2 
% end
         meth ComplexSeq(L $) % --> Vars#Factors#Actions
            Seq = {New SeqManager init}
         in
            {List.forAllInd L
             proc {$ N CModel}
                C="C"#N
                M="M"#N
                R=Rules, get(CModel $)
                Vars Factor Action
             in 
                {Simple C M R ?Vars ?Factor ?Action}?
                {Seq putVars(C|M|Vars)}
                {Seq putFactor(Factor)}
                {Seq putAction(Action)}
             end}    
            local
               N = {Length L}
               fun {MakeFinalConcatenation Var}
                  Var#"="#{ForThread 2 N 1
                           fun {$ VS I} VS#"#"#Var#I end
                           Var#"1"}
               end
            in
               {Seq putAction({MakeFinalConcatenation "C"})}
               {Seq putAction({MakeFinalConcatenation "M"})}
            end
            {Seq getVars($)}#{Seq getFactors($)}#{Seq getActions($)}
         end
         meth CreateSimple(C M CModel ?Vars ?Factor ?Action)
            case CModel
            of '?'(SubModel) then
               R = Rules, get(SubModel $) in
               {Optional C M R ?Vars ?Factor ?Action}
            [] '*'(SubModel) then
               R = Rules, get(SubModel $) in
               {Star C M R ?Vars ?Factor ?Action}
            [] '+'(SubModel) then
               R = Rules, get(SubModel $) in
               {Plus C M R ?Vars ?Factor ?Action}
            else
               R = Rules, get(CModel $) in
               {Simple C M R ?Vars ?Factor ?Action}   
            end
         end
         %%
         %% element rules
         %%
         meth createElementRule(E)
            case E.content
            of 'EMPTY'  then
               Rules, PutElementRule(E "" "unit" "'EMPTY'")
            [] 'ANY'    then
               Rules, PutElementRule(E "ANY(X)" "X" "'ANY'")
            [] 'PCDATA' then
               Rules, PutElementRule(E "PCDATA_OPT(VS)" "VS" "'PCDATA'")
            [] mixed(Elems) then
               Rules, CreateMixed(E "PCDATA" Elems)
            [] children(Spec) then
               case Rules, get(Spec $) %% choice,seq,?,+,*,'<ELEM>'
               of element(R) then
                  Rules, PutElementRule(E {Elem R}#"(Y)" "Y" "'"#R#"'")
               [] aux(R) then
                  Rules, PutElementRule(E {Aux R}#"(?C ?M)" "C" "M")
               end
            end
         end
         meth CreateMixed(E PCDATA Elems)
            FElems = case Elems of [Element] then {Elem Element}#"($)"
                     elsecase Rules, get('choice'(Elems) $)
                     of element(R) then {Elem R}#"($)"
                     [] aux(R)     then {Aux R}#"($ _)"   
                     end
         in
            Rules, PutElementRule(E "C=mixed: ( "#PCDATA#"($) // "#FElems#" )*"
                                  "C" "'mixed'")
         end
         meth PutElementRule(E F C M)
            N = E.name 
            Head   = {Elem N}#"($)"
            Factor = {Tok N}#"(E) "#F#" END"
            Action = {@writeAction E C M}
         in
            Rules, putRule({WriteSyn Head [{WriteAlt nil [Factor] [Action]}]})
         end
      end
   end

   %%
   %% exports
   %%
   
   fun {WriteTerminalRule ElemName}
      Head = {Term ElemName}#"(?X)"
      Body = "!X="#{Elem ElemName}#"($)"#"\n"
   in
      {WriteSyn Head [Body]}
   end
   
   fun {MakeWriteAction SemClass PrefixVS}
      if SemClass==false then
         fun {$ E C M}
            ""      #"E.content = "#C#"\n"#
            "      "#"E.model   = "#M#"\n"#
            "      "#"{self '"#PrefixVS#{ToString E.name}#"'(E $)}"
         end
      else
         fun {$ E C M}
            ""      #"E.content = "#C#"\n"#
            "      "#"E.model   = "#M#"\n"#
            "      "#SemClass#", '"#PrefixVS#{ToString E.name}#"'(E $)"
         end
      end
   end
   
   fun {WriteElementRules Elements WriteAction}
      Rs = {New Rules init(writeAction:WriteAction)} in
      {ForAll Elements proc {$ E} {Rs createElementRule(E)} end}
      {Enum {Rs getRules($)} "\n"} 
   end

   fun {WriteSpecialRule RuleName Elements} % matching any 'START_...' token
      Head = RuleName#"($)"
      As   = {Map Elements fun {$ E} {Tok E.name}#"($)"#"\n" end}
   in
      {WriteSyn Head As}
   end
 
end
