%%
%% expat emulator
%%
%% adopted from expat.oz (Denys Duchier) 
%%

functor

import
   InputSource at 'x-ozlib://duchier/op/InputSource.ozf'
   XmlScanner  at 'x-ozlib://mathweb/xmlparser/XmlScanner.ozf'
   XVS         at 'XVS.ozf'
   
export
   'class' : Parser

define
   
   local
      fun {ReadLoop Source}
         case {Source get($)} of nil then nil
         [] VS then VS#{ReadLoop Source}
         end
      end
   in
      proc {SourceToVs Source ?Vs}
         Vs = {ReadLoop Source}
         try {Source close()}
         catch _ then skip end
      end
   end

   fun {RefToVS Ref}
      case Ref of '\&'(RefName)  then "\&"#RefName#";"
      else
         raise expatEmulator(unexpected(Ref)) end
      end
   end
   fun {AttribValsToVS AList}
      {Map AList fun {$ Name#Val} Name#{XVS.map Val RefToVS} end}
   end
   
   VS2A = VirtualString.toAtom

   class Parser
      attr events
      meth init(Source)
         events<-{XmlScanner.document.scanVS {SourceToVs Source}}
      end
      %%
      meth initFromFile(File)
         Parser,init({New InputSource.file init(File)})
      end
      meth initFromFiles(Files)
         Parser,init({New InputSource.files init(Files)})
      end
      meth initFromURL(URL)
         Parser,init({New InputSource.url init(URL)})
      end
      meth initFromURLs(URLs)
         Parser,init({New InputSource.urls init(URLs)})
      end
      meth initFromVs(Vs)
         Parser,init({New InputSource.vs init(Vs)})
      end
      meth initFromString(Str)
         Parser,init({New InputSource.string init(Str)})
      end
      %%
      meth close()
         events<-unit
      end
      %%
      meth GetEvent($)
         case @events
         of unit then unit
         [] nil  then unit
         [] (Tok#Val)|T then
            events<-T
            case Tok
            of startElement then Tag#AList=Val in
               startElement(Tag {AttribValsToVS AList})
            [] endElement   then /* Atom         */ endElement(Val)
            [] characters   then /* VirtualString*/ characters(Val)
            [] 'PI'         then
               'PI'(TargetVS CharsVS) = Val in
               processingInstruction({VS2A TargetVS} CharsVS)
            [] 'CDATA'      then
               events<-(characters#Val)|(endCDATA#unit)|@events
               startCDATA()
            [] startCDATA   then startCDATA()
            [] endCDATA     then endCDATA()
            [] entityRef    then characters('\&'#Val#';')
            [] '<!DOCTYPE'  then /* N#ExtID          */ Parser, GetEvent($)
            [] 'ELEMENT'    then /*'ELEMENT'(N Model)*/ Parser, GetEvent($)
            [] 'ATTLIST'    then /*'ATTLIST'(N Defs) */ Parser, GetEvent($)
            [] 'ENTITY'     then /*'\&'(N Def)       */ Parser, GetEvent($)
               /*                  '%'(N Def)        */
            [] 'NOTATION'   then /*'NOTATION'(N ID)  */ Parser, GetEvent($)
            [] '>'          then /* unit             */ Parser, GetEvent($)
            [] 'EOF'        then unit
            end
         end
      end
      %%
      meth parse()
         {self startDocument()}
         Parser,Loop()
         {self endDocument()}
      end
      %%
      meth Loop()
	 case Parser,GetEvent($)
	 of unit then skip
	 [] E then {self E} Parser,Loop() end
      end
      %%
      meth startDocument()              skip end
      meth endDocument()                skip end
      meth startElement(_ _)            skip end
      meth endElement(_)                skip end
      meth characters(_)                skip end
      meth processingInstruction(_ _)   skip end
      meth comment(_)                   skip end
      meth startCDATA()                 skip end
      meth endCDATA()                   skip end
      meth default(_)                   skip end
      /*
      meth startNamespaceDeclScope(_ _) skip end
      meth endNamespaceDeclScope(_)     skip end
      */
   end
end

