%%%
%%% Authors:
%%%   Denys Duchier <duchier@ps.uni-sb.de>
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Denys Duchier, 1999
%%%   Andreas Franke, 1999
%%%
%%% Last change:
%%%   $Date: 2000/12/07 13:49:12 $ by $Author: afranke $
%%%   $Revision: 1.4 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% an expat interface that emulates a Gump scanner for XML
%% (to be used as token source for a Gump parser)
%%

functor
   
import
   System(show printInfo)

\ifdef EXPAT
   X('class':BaseParser) at 'x-ozlib://duchier/xml/expat/Processor.ozf'
\else
   Y('class':BaseParser) at 'x-ozlib://mathweb/xmlparser/ExpatEmulator.ozf'
\endif
   Z('class': FIFO)      at 'FIFO.ozf'
   
export
   'class':  Scanner
   stepwise: StepScanner
   
define
  
   START      = 'START'
   END        = 'END'
   CHARACTERS = 'characters'
   
   class Scanner
      from BaseParser
         
      attr
         textModes:    unit % for each element: is PCDATA allowed?
         attDefaults:  unit % for each element: default values for attributes
         tokens:       unit % for each element: startElement token
         defaultToken: unit % startElement token for unknown elements
         %% namespace
         ns:      unit
         nsStack: nil
         %% current text mode
         text:      true
         textStack: nil
         %% token queue (FIFO)
         queue: unit
         %% 
         debug: false
         
      meth xmlInit(elements: Spec        <= unit
                   default:  DefaultTok  <= START
                   debug:    Debug       <= false)
         tokens       <- {CondSelect Spec 'token'  unit}
         textModes    <- {CondSelect Spec text     unit}
         attDefaults  <- {CondSelect Spec defaults unit} 
         defaultToken <- DefaultTok
         queue        <- {New FIFO init}
         debug        <- Debug
      end 
   
      meth DebugPrintInfo(VS)
         if @debug then
            {System.printInfo VS}
         end
      end
      %%
      meth close
         BaseParser, close
         queue <- unit
      end

      %%
      %% token stream
      %%
      meth putToken(T V)
         {@queue put(T#V)}
      end
      meth putToken1(T)
         Scanner, putToken(T unit)
      end
      meth getToken(?T ?V)
         case {@queue get($)} of Tok#Val then T=Tok V=Val
            if @debug then {System.show T#V} end
         [] false then
            %% This case appears only once!
            %% Initially, the token queue is empty.
            %% When the Gump parser begins a parse,
            %% it first calls putToken(StartToken '#').
            %% Then it calls getToken(?T ?V) repeatedly, until T=='EOF'.
            %% Thus the first call of this method returns the StartToken,
            %% and the second call leads to this case.
            %% Now we tokenize the complete document.
            BaseParser, parse()
            %% This includes a call of endDocument()
            %% which appends 'EOF' to the token stream.
            Scanner, getToken(?T ?V)
         end
      end
   
      meth initTokenStream
         {@queue init}
      end
      meth closeTokenStream
         Scanner, putToken1('EOF')
      end
      
      %%
      %% %%%%%%%%%%%%%%%%%%%%%%%%%%%% events %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%
      meth startDocument()
         ns <- ns
         %% Scanner, initTokenStream % does not work together with Gump
      end
      meth endDocument()
         Scanner, closeTokenStream
      end
      %%
      meth startElement(Tag AList)
         Scanner, DebugPrintInfo("(")
         Attribs = {Record.subtract
                    {AdjoinList {CondSelect @attDefaults Tag Tag} AList}
                    xmlns} %% bloody hack to avoid clutter e.g. in OMOBJs
         Val     = element(name:Tag ns:@ns attribs:Attribs content:_ model:_)
      in
         Scanner, putToken({CondSelect @tokens Tag @defaultToken} Val)
         Scanner, PushText({CondSelect @textModes Tag @text})
      end
      meth endElement(Tag)
         Scanner, DebugPrintInfo(")") 
         Scanner, putToken(END Tag)
         Scanner, PopText
      end
      meth characters(Vs)
         Scanner, DebugPrintInfo(".")
         case @text of false then skip
         [] true    then Scanner, putToken(CHARACTERS Vs)
            /*
         [] 'parse' then Scanner, ParseChars({VirtualString.toString Vs})
            */
         end
      end
      %%
      meth processingInstruction(Target Data) Scanner, DebugPrintInfo('?') end
      meth comment(_)                         Scanner, DebugPrintInfo('c') end
      meth startCDATA()                       Scanner, DebugPrintInfo('[') end
      meth endCDATA()                         Scanner, DebugPrintInfo(']') end
      meth default(_)                         Scanner, DebugPrintInfo('_') end
      %%
      meth startNamespaceDeclScope(Prefix URI)
         Scanner, PushNS({AdjoinAt @ns Prefix URI})
         Scanner, DebugPrintInfo(':')
      end
      meth endNamespaceDeclScope(_)
         Scanner, PopNS
         Scanner, DebugPrintInfo(';')
      end
      %%
      %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      %%
      
      %% basic stack operations (text)
      meth PushText(X)
         textStack <- @text|@textStack
         text      <- X
      end
      meth PopText
         case @textStack of H|T then
            textStack <- T
            text      <- H
         end
      end
      %% basic stack operations (namespace)
      meth PushNS(X)
         nsStack <- @ns|@nsStack
         ns      <- X
      end
      meth PopNS
         case @nsStack of H|T then
            nsStack <- T
            ns      <- H
         end
      end
      
      %%
      %% parse character-strings (works, but is not needed anymore)
      %%      
      /*
      meth ParseChars(S)
         case S of nil then skip
         [] C|Sr then
            %% 'alphas'#VS 'digits'#VS 'spaces'#VS 'punct'#C 'other'#C
            Rest = case {Char.type C}
                   of lower then Scanner, ProcessAlpha(C Sr $ capital:false)
                   [] upper then Scanner, ProcessAlpha(C Sr $ capital:true)
                   [] digit then Scanner, ProcessDigit(C Sr $)
                   [] space then Scanner, ProcessSpace(C Sr $)     
                   [] punct then Scanner, putToken(punct C)  Sr
                   [] other then Scanner, putToken(other C)  Sr
                   end
         in
            Scanner, ParseChars(Rest)
         end
      end
      meth ProcessAlpha(C Sr ?Rest capital:Cap)
         S = C|{List.takeDropWhile Sr Char.isAlpha $ ?Rest} in
         Scanner, putToken(alphas S)
      end
      meth ProcessDigit(C Sr ?Rest)
         S = C|{List.takeDropWhile Sr Char.isDigit $ ?Rest} in
         Scanner, putToken(digits S)
      end
      meth ProcessSpace(C Sr ?Rest)
         S = C|{List.takeDropWhile Sr fun {$ X} X==C end $ ?Rest} in
         Scanner, putToken(spaces S)
      end
      */
   end

   %%
   %% stepwise tokenization (experimental)
   %%
   class StepScanner
      from Scanner
         
      meth init(...) = M
         Scanner, M
         Scanner, startDocument
      end
      
      meth getToken(?T ?V)
         case {@queue get($)} of Tok#Val then T=Tok V=Val
            if @debug then {System.show T#V} end
         [] false then
            Scanner, step()
            StepScanner, getToken(?T ?V)
         end
      end
   end
   
end  
