%%%
%%% This is a first implementation of "eXtended virtual strings" (Xvs)
%%% based on the extended virtual strings known from `Error' system module.
%%% Use it or forget it...:-)
%%%
%%% Some features of this implementation are based on
%%%   - the `VirtualString' base module
%%%   - the `Error' system module
%%% 
%%% Creator of this file:
%%%   Andreas Franke <afranke@ags.uni-sb.de>

%%
%%    <extended virtual string> ::=
%%         '#'(<extended virtual string> ... <extended virtual string>)
%%       | <simple-vs>            
%%       | <x-element>
%%
%%    <simple-vs> ::= atom | int | float | string | byteString
%%
%%    <x-element> ::= any oz-value V with the following properties:
%%                    - {IsDet V}
%%                    - V is not a <simple-vs>
%%                    - if {IsRecord V} then
%%                        {Label V}\='|' andthen
%%                        {Label V}\='#'
%%                      end
%%
%%    Recommendation:
%%
%%    <x-element> ::= any stateless non-atom record R
%%                      with label not in ['|' '#'],
%%                    some tuples should only be used with the semantics
%%                    defined in the `Error' system module:
%%                      oz(M), pn(M), pos(F L C), apply(A Xs), list(Xs Sep) 
%%

functor

export
   all:               XvsAll
   some:              XvsSome
   map:               XvsMap
   mapCond:           XvsMapCond
   mapUnless:         XvsMapUnless
   dropAll:           XvsDropAll
   filterOut:         XvsFilterOut
   filter:            XvsFilter
   text:              XvsText
   elements:          XvsElements
   normalize:         XvsFlatten
   normalizeWithTail: XvsFlattenWithTail
   
define
   /*
   fun {IsStateless X}
      %% must be reimplemented!!! 
      %% Pickle.pack crashes the emulator :(
      try {Pickle.save X "/dev/null"} true
      catch _ then false end
   end
   fun {IsExtension X}
      {IsDet X} andthen
      {IsRecord X} andthen
      {Member {Label X} ['|' '#']}==false andthen
      {IsStateless X}
   end
   */
   
   fun {IsSimple1 X}
      case {Value.type X}
      of atom then true
      [] int then true
      [] float then true
      [] byteString then true
      elsecase X of '|'(_ _) then true
      else false
      end
   end
   fun {IsSimple2 X}
      case X of '#'(...) then {Record.all X IsSimple2}
      else {IsSimple1 X} end
   end
   
%   IsSimple = IsSimple1          
   IsSimple = IsSimple2
%   IsSimple = IsVirtualString 
   
   %%
   %% all & some 
   %%

   fun {XvsAll X Cond}
      fun {Test Y}
         case Y of '#'(...) then {Record.all Y Test}
         elseif {IsSimple Y} then true 
         else {Cond Y} end
      end
   in
      {Test X}
   end
   fun {XvsSome X Cond}
      fun {Test Y}
         case Y of '#'(...) then {Record.some Y Test}
         elseif {IsSimple Y} then false 
         else {Cond Y} end
      end
   in
      {Test X}
   end

   %%
   %% map
   %%

   fun {XvsMap X F}
      fun {Map Y}
         case Y of '#'(...) then  {Record.map Y Map} 
         elseif {IsSimple Y} then Y
         else {F Y} end
      end
   in
      {Map X}
   end
   fun {XvsMapCond X F Cond}
      fun {CondF X}
         if {Cond X} then {F X} else X end
      end
   in
      {XvsMap X CondF}
   end
   fun {XvsMapUnless X F Cond}
      fun {CondF X}
         if {Cond X}==false then {F X} else X end 
      end
   in
      {XvsMap X CondF}
   end

   %%
   %% filter 
   %%

   local
      fun {Drop _} '' end
   in
      fun {XvsDropAll X}
         {XvsMap X Drop} 
      end
      fun {XvsFilterOut X Cond}
         {XvsMapCond X Drop Cond}
      end
      fun {XvsFilter X Cond}
         {XvsMapUnless X Drop Cond}
      end
   end
   
   %%
   %% normalize to X1#(X2#(X3#...(Xn#'')...))
   %%
   
   local
      fun {FlattenAll I X Cond Tail}
         if I>0 then {FlattenAll I-1 X Cond {XvsFlattenWithTail X.I Cond Tail}}
         else Tail
         end
      end
      fun {True _} true end
   in
      fun {XvsFlattenWithTail X Cond Tail}
         case X of '#'(...) then {FlattenAll {Width X} X Cond Tail}
         elseif {Cond X} then X#Tail
         else Tail end
      end
      fun {XvsFlatten X}
         {XvsFlattenWithTail X True ''}
      end
      fun {XvsText X}
         {XvsFlattenWithTail X IsSimple ''}
      end
   end

   %%
   %% convert to list & filter
   %%
   
   local
      fun {ToListAll I X Cond Tail}
         if I>0 then {ToListAll I-1 X Cond {XvsToListWithTail X.I Cond Tail}}
         else Tail
         end
      end
      fun {IsElement X} {IsSimple X}==false end
   in
      fun {XvsToListWithTail X Cond Tail}
         case X of '#'(...) then {ToListAll {Width X} X Cond Tail}
         elseif {Cond X} then X|Tail
         else Tail end
      end
      fun {XvsElements X}
         {XvsToListWithTail X IsElement nil}
      end
   end
end
 
%%%
%%% Note:
%%%
%%% There can be built several different concrete implementations
%%% of "mixed" values on top of this, e.g.
%%%
%%% - an Xvs containing Oz-names and a corresponding record or dictionary
%%%   mapping these names to x-elements
%%%
%%% - an Xvs containing records of the form lookup(key:42) and a corresponding
%%%   record or dictionary
%%%
%%% - ....
%%%

/*
declare X = "string"#el(1)#'atom'#42#3.14#optional(2)#{ByteString.make "BS"}
{Browse X}
{Browse {Xvs.elements X}}
{Browse {Xvs.text X}}
{Browse {VirtualString.toAtom {Xvs.text X}}}
{Browse {Xvs.filter X fun {$ E} {Label E}\=optional end}}
{Browse {Xvs.normalize X}}
{Browse {Xvs.all X fun {$ E} {Label E}\=optional end}}
{Browse {Xvs.some X fun {$ E} {Label E}\=optional end}}
*/

