%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 1999
%%%
%%% Last change:
%%%   $Date: 2000/01/16 06:37:44 $ by $Author: afranke $
%%%   $Revision: 1.2 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% XML parser interface
%%

functor
   
import
   Error(registerFormatter) Open(file) Type
   
   GumpScannerEmulator
   at 'x-ozlib://mathweb/xmlparser/GumpScannerEmulator.ozf'
   
export
   setDebug: SetDebug % proc{$ FlagB}
   
   parse: Parse       % proc{$ ScannerInit ParserModule StartSym(?Val) ?Accept}
   
   parseString: ParseString % proc{$ Str   ParserModule StartSym(?Val) ?Accept}
   parseVS:     ParseVS     % proc{$ Vs    ...                             ...}
   parseFile:   ParseFile   % proc{$ FileName...                           ...}
   parseFiles:  ParseFiles  % proc{$ FileNames...                          ...}
   parseURL:    ParseUrl    % proc{$ Url   ...                             ...}
   parseURLs:   ParseUrls   % proc{$ Urls  ...                             ...}

   link: LinkParser     % fun{$ GenericParserModule SemModule} --> ParserModule
   
define
   
   DebugFlag = {NewCell false}

   proc {SetDebug FlagB}
      {Type.ask.bool FlagB}
      {Assign DebugFlag FlagB}
   end
   
   %%
   %% main function
   %%
   proc {Parse ScannerInit ParserModule ParseTuple ?Accept}
      if ParserModule.isGeneric\=false then
         {Exception.raiseError xmlParser(generic ParserModule)}
      else
         ScannerObj = {New GumpScannerEmulator.'class' ScannerInit}
         ParserObj  = {New ParserModule.'class' init(ScannerObj)}
      in
         {ScannerObj xmlInit(elements: ParserModule.elements
                             debug:    {Access DebugFlag})}
         {ParserObj parse(ParseTuple ?Accept)}
         {ScannerObj close()}
      end
   end
   
   %%
   %% specialized versions of Parse
   %%
   local
      fun {MkParse InitLabel}
         proc {$ Input ParserModule ParseTuple ?Accept}
            {Parse InitLabel(Input) ParserModule ParseTuple ?Accept}
         end
      end
   in
      ParseString = {MkParse initFromString}
      ParseVS     = {MkParse initFromVs}
      ParseFile   = {MkParse initFromFile}
      ParseFiles  = {MkParse initFromFiles}
      %%ParseUrl    = {MkParse initFromURL} % see below
      ParseUrls   = {MkParse initFromURLs}
   end

   %%
   %% workaround for gc bug (temporarily)
   %%
   proc {ParseUrl Url ParserModule ParseTuple ?Accept}
      BS in
      local
         F  = {New Open.file init(url:Url)} in
         BS = {ByteString.make {F read(list:$ size:all)}}
         {F close}
      end
      {Parse initFromVs(BS) ParserModule ParseTuple ?Accept}
   end

   %%
   %% dynamic linking of the semantic module
   %%
   fun {LinkParser ParserModule SemModule}
      if ParserModule.isGeneric\=true then
         {Exception.raiseError xmlParser(nonGeneric ParserModule)} unit
      else
         'export'('class':   class $
                                from
                                   ParserModule.'class'
                                   SemModule.'class'
                             end
                  elements:  ParserModule.elements
                  isGeneric: false)
      end
   end
   
   %% %%%%%%%%%%%%%%%%%%%%%%%%% error formatter %%%%%%%%%%%%%%%%%%%%%%%%%
   {Error.registerFormatter xmlParser
    fun {$ E}
       T = 'xml parser error' 
    in 
       case E
       of xmlParser(nonGeneric ParserModule) then 
          error(kind:T msg:'Non-generic parser module cannot be linked' 
                items:[line(oz(ParserModule))])
       [] xmlParser(generic ParserModule) then
          error(kind:T msg:'Generic parser module has to be linked before use' 
                items: [line(oz(ParserModule))])
       else
          error(kind:T items:[line(oz(E))])
       end 
    end}
   %% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

end



