%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 1999
%%%
%%% Last change:
%%%   $Date: 2000/01/16 06:08:22 $ by $Author: afranke $
%%%   $Revision: 1.1 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% text2functor.oz
%%

functor
   
import
   Application(getArgs exit)
   System(showInfo showError)
   Open(file)
   Pickle(save)
   
prepare
   %%
   %% string suffix
   %%
   fun {IsSuffix S Suffix}
      {List.isPrefix {Reverse Suffix} {Reverse S}}
   end
   fun {DropSuffix S Suffix}
      if {IsSuffix S Suffix}==false then S
      else
         {Reverse {List.drop {Reverse S} {Length Suffix}}}
      end
   end
   
   %%
   %% take a VirtualString and turn it into a functor
   %% that exports the corresponding ByteString / String
   %%
   fun {MakeFunctor VS}
      BS = {ByteString.make VS} in
      functor $
      export
         byteString: BS
         string:     S
      define
         S = {ByNeed fun {$} {ByteString.toString BS} end}
      end
   end
   
define
   AppName = 'text2functor'
   
   %%
   %% get args
   %%
   Spec = record(input(single char:&i type:string optional:true)
                 output(single char:&o type:string optional:true)
                 quiet(single char:&q type:bool default:false)
                )
   Args
   try
      Args = {Application.getArgs Spec}
   catch error(ap(usage S)...) then
      {System.showError AppName#": "#S}
      {Application.exit 1}
   end

   %%
   %% compute list of pairs InputFile#OutputFile
   %%
   fun {GetOutput Default}
      {CondSelect Args output Default}
   end
   
   DefaultInput  = 'stdin'
   DefaultOutput = {GetOutput 'stdout'}
   
   Inputs = if {HasFeature Args input} then [Args.input]
            else Args.1 end
   
   if {HasFeature Args output} andthen {Length Inputs}>1 then
      {System.showError AppName#": "#
       "output file specified, but multiple inputs found"}
      {Application.exit 1}
   end
   
   IOs = case Inputs of nil then [DefaultInput#DefaultOutput]
         else
            fun {MakeIO X}
               X#{GetOutput {DropSuffix X ".ozg"}#".ozf"}
            end
         in
            {Map Inputs MakeIO}
         end
   
   %%
   %% load a text file
   %% 
   proc {LoadVS File ?VS}
      F = {New Open.file init(name:File)} in
      {F read(list:?VS size:all)}
      {F close}
   end

   %%
   %% do it!
   %%
   {ForAll IOs proc {$ I#O}
                  {Pickle.save {MakeFunctor {LoadVS I}} O}
                  if Args.quiet==false andthen O\='stdout' andthen
                     {HasFeature Args output}==false
                  then
                     {System.showInfo "Wrote "#O}
                  end
               end}
   {Application.exit 0}
end









