%%%
%%% XML parser demo application
%%%

functor
   
import
   Application Browser Inspector Module Pickle Property System
   
   XmlParser  at 'x-ozlib://mathweb/xmlparser/XmlParser.ozf'
   SemDefault at 'x-ozlib://mathweb/xmlparser/SemDefault.ozf'
   
define
   
   %%
   %% auxiliuary timer functions
   %%
   fun {GetTimes}
      {Record.filter {Property.get 'time'} IsInt}
   end
   fun {DiffTimes T1 T2}
      {Record.zip T2 T1 Number.'-'}
   end
   proc {ShowTimes Times}
      {System.showInfo ''}         
      {ForAll [system user total]
       proc {$ F} {System.showInfo F#':\t'#Times.F} end}
      {System.showInfo ''}
   end

   %%
   %% spec & args
   %%
   AppName = 'xmlread'
   Spec    = record(
                savepickle(single char:[&p &o] type:string default:false)
                compress(single char:&z type:int(min:0 max:9) default:9)
                mode(single type:atom(show browse inspect write quiet)
                     default:write)
                show(   char:&s alias:mode#show)
                browse( char:&b alias:mode#browse)
                inspect(char:&i alias:mode#inspect)
                write(  char:&w alias:mode#write)
                echo(   char:&e alias:mode#write)
                quiet(  char:&q alias:mode#quiet)
                debug(single char:&d type:bool default:false)
                %% browser options:
                strings(single type:bool default:true)
                virtualstrings(single type:bool default:true)
                vs(alias:virtualstrings)
                
                parser(single type:string optional:false)
                semantics(single type:string default:unit)
                defaultsemantics(single type:string default:unit)
                
                root(single type:string default:unit)
                startsymbol(single type:atom default:unit)
                any(        single alias:[startsymbol#any])
                flat(       single alias:[startsymbol#flat])
                misc(       single alias:[startsymbol#misc])
                )
   
   %%
   %% shortcuts for some example parsers 
   %%
   local
      ParserDir = "x-ozlib://mathweb/xml/parser/"
      SemDir    = "x-ozlib://mathweb/xml/parser/"
      %%
      local    
         fun {GetParser SpecRec}
            if {HasFeature SpecRec parser} then SpecRec.parser
            else ParserDir#{Label SpecRec}#".parser.ozf" end
         end
         fun {GetSem SpecRec}
            if {HasFeature SpecRec sem} then SpecRec.sem
            else SemDir#{Label SpecRec}#".sem.ozf" end
         end
         fun {GetRoot SpecRec}
            {CondSelect SpecRec root {Label SpecRec}}
         end
      in
         fun{MakeAlias SpecRec}
           L = {Label SpecRec} in
           L(single alias: [parser           # {GetParser SpecRec}
                            defaultsemantics # {GetSem    SpecRec}
                            root             # {GetRoot   SpecRec}])
         end
      end
   in
      local
         Specs =
         ['omobj'(root:"OMOBJ")
          'omdoc'(root:"omdoc")
          'omcd'(root:"CD")
          'cdgroup'(root:"CDGroup")
          'omcdsig'(root:"CDSignatures")
          'xsl'(root:"xsl:stylesheet")
          'xslstylesheet'(parser:ParserDir#"xsl.parser.ozf"
                          sem:SemDir#"xsl.sem.ozf" root:"xsl:stylesheet")
          'xsltransform'(parser:ParserDir#"xsl.parser.ozf"
                         sem:SemDir#"xsl.sem.ozf" root: "xsl:transform")
          'kqml'(root:"kqml")
         ]
      in
         ShortCuts = {List.toTuple record {Map Specs MakeAlias}}
      end
   end
   
   Args
   try
      Args = {Application.getArgs {Tuple.append ShortCuts Spec}}
   catch error(ap(usage S)...) then
      {System.showError AppName#': '#S}
      {Application.exit 1}
   end
   
   %%
   %% input parameters of `Parse'
   %%
   Inputs = case Args.1 of nil then ['/dev/stdin']
            [] Us then {Map Us VirtualString.toAtom} end
   
   %% load parser & semantics module
   ModMan = {New Module.manager init}
   
   SModuleUrl = case Args.semantics
                of unit then Args.defaultsemantics
                [] Url  then Url
                end
   {System.printInfo "semantics module: "}
   if SModuleUrl==unit
   then {System.showInfo "<default>"}
   else {System.showInfo SModuleUrl}
   end
   
   PModule = {ModMan link(url:Args.parser $)}
   SModule = if SModuleUrl==unit
             then SemDefault
             else {ModMan link(url:SModuleUrl $)}
             end
   
   %% link generic module with semantics module
   ParserModule = if Args.semantics\=unit orelse PModule.isGeneric==true
                  then {XmlParser.link PModule SModule}
                  else PModule
                  end

   %% startsymbol
   Startsymbol = case Args.startsymbol of unit then
                    case Args.root of unit then any
                    [] Elem then {VirtualString.toAtom "start_"#Elem}
                    end
                 [] Sym then Sym
                 end
   {System.showInfo "start symbol    : "#Startsymbol}
 
   {XmlParser.setDebug Args.debug}

   {Property.put 'errors.width' 1000}
   {Property.put 'errors.depth' 1000}
   
   %%
   %% output parameters of `Parse'
   %%
   ?Val
   ?Accept

   %%
   %% parser call
   %%
   T1 = {GetTimes}
   case Inputs of [Url]
   then {XmlParser.parseURL  Url    ParserModule Startsymbol(?Val) ?Accept}
   else {XmlParser.parseURLs Inputs ParserModule Startsymbol(?Val) ?Accept}
   end
   T2 = {GetTimes}
   {ShowTimes {DiffTimes T1 T2}}
   
   if ?Accept then
      {System.showInfo 'accepted'}
   else
      {System.showInfo 'rejected'}
      {Application.exit 1}
   end
   
   case Args.savepickle of false then
      case Args.mode
      of show   then {System.show Val}
      [] browse then
         if Args.strings then
            {Browser.object option(representation strings:true)}
         end
         if Args.virtualstrings then
            {Browser.object option(representation virtualStrings:true)}
         end
         {Browser.browse Val}
         {Thread.suspend {Thread.this}}
      [] inspect then {Inspector.inspect Val}
         {Thread.suspend {Thread.this}}
      [] write then
         if {IsVirtualString Val}
         then {System.printInfo Val}
         else {System.show Val}
         end
      [] quiet then skip
      end
   [] FileName then {Pickle.saveCompressed Val FileName Args.compress}
   end
   {Application.exit 0}
end

